@@ -85,7 +85,7 @@ int ibv_read_sysfs_file(const char *dir, const char *file,
 	if (asprintf(&path, "%s/%s", dir, file) < 0)
 		return -1;
 
-	fd = open(path, O_RDONLY);
+	fd = open(path, O_RDONLY | O_CLOEXEC);
 	if (fd < 0) {
 		free(path);
 		return -1;
